"""Decorator Functions."""
import functools
import time
from typing import Any, Callable, Dict, List, Tuple, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
from prettytable import PrettyTable
from rich.pretty import pprint

from common_utils.core.common import seed_all

seed_all(1992)

DataTypes = Union[List[int], Dict[int, int], None]
#  callable that takes any number of arguments and returns any value.
F = TypeVar("F", bound=Callable[..., Any])


# pylint: disable=invalid-name
def data_factory(data_type: str, n: int) -> DataTypes:
    if data_type == "array":
        return list(range(n))
    if data_type == "dict":
        return {i: i for i in range(n)}
    if data_type is None:
        return None
    raise ValueError(f"Invalid data_type: {data_type}")


def time_complexity(
    data_type: str, repeat: int = 1, plot: bool = False
) -> Callable[[Callable[..., Any]], Callable[..., Tuple]]:
    def decorator(func: Callable[..., Any]) -> Callable[..., Tuple]:
        def wrapper(n_sizes: List[int], *args: Any, **kwargs: Dict[str, Any]) -> Tuple:
            avg_times = []
            median_times = []
            best_times = []
            worst_times = []

            for n in n_sizes:
                # create a list of n elements
                data_structure = data_factory(data_type, n)
                # note array is created outside the loop
                runtimes = []
                for _ in range(repeat):
                    start_time = time.perf_counter()

                    # pylint: disable=expression-not-assigned,line-too-long
                    func(n, data_structure, *args, **kwargs) if data_type else func(
                        n, *args, **kwargs
                    )  # <--- this is where it calls the function with n or data_structure as argument
                    end_time = time.perf_counter()
                    runtimes.append(end_time - start_time)

                avg_times.append(np.mean(runtimes))
                median_times.append(np.median(runtimes))
                best_times.append(np.min(runtimes))
                worst_times.append(np.max(runtimes))

            if plot:
                plt.figure(figsize=(10, 6))
                plt.plot(n_sizes, avg_times, "o-", label="Average")
                plt.plot(n_sizes, median_times, "o-", label="Median")
                plt.plot(n_sizes, best_times, "o-", label="Best")
                plt.plot(n_sizes, worst_times, "o-", label="Worst")
                plt.xlabel("Size of Input (n)")
                plt.ylabel("Execution Time (s)")
                plt.legend()
                plt.grid(True)
                plt.title(f"Time Complexity of {func.__name__}")
                plt.show()

            return n_sizes, avg_times, median_times, best_times, worst_times

        return wrapper

    return decorator


def timer(func: F) -> F:
    """Timer decorator."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Dict[str, Any]) -> Any:
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time

        # Create a table to display the results
        table = PrettyTable()
        table.field_names = ["Function Name", "Seconds", "Minutes", "Hours"]
        table.add_row(
            [
                func.__name__,
                f"{elapsed_time:.4f}",
                f"{elapsed_time / 60:.4f}",
                f"{elapsed_time / 60 / 60:.4f}",
            ]
        )

        pprint(table)
        return result

    return wrapper


@time_complexity(data_type="array", repeat=10, plot=True)
def list_access(n: int, array) -> None:
    _ = array[n // 2]


@time_complexity(data_type="array", repeat=10, plot=True)
def list_append(n: int, array) -> None:
    array.append(n)


@time_complexity(data_type="array", repeat=10, plot=True)
def list_insert(n: int, array) -> None:
    array.insert(0, n)


@time_complexity(data_type="array", repeat=10, plot=True)
def list_search(n: int, array) -> None:
    _ = n in array


# @time_complexity(repeat=10, plot=True)
# def for_loop(n: int, array) -> None:
#     for i in range(n):


@time_complexity(data_type="dict", repeat=10, plot=True)
def dict_set(n: int, dict_) -> None:
    dict_[n] = n


@time_complexity(data_type="dict", repeat=10, plot=True)
def dict_search(n: int, dict_) -> None:
    _ = n in dict_


@time_complexity(data_type=None, repeat=10, plot=True)
def for_loop(n: int) -> None:
    for _ in range(n):
        pass


@timer
def add_two_arrays(array_1: np.ndarray, array_2: np.ndarray) -> np.ndarray:
    """Add two arrays together."""
    return array_1 + array_2


# list_access(range(1000000, 10000001, 1000000))
# list_append(range(1000000, 10000001, 1000000))
# list_insert(range(1000000, 10000001, 1000000))
# list_search(range(1000000, 10000001, 1000000))

dict_set(range(1000000, 10000001, 1000000))
dict_search(range(1000000, 10000001, 1000000))
for_loop(range(1000000, 10000001, 1000000))


# array_1 = np.random.randint(0, 100, size=(10000, 10000))
# array_2 = np.random.randint(0, 100, size=(10000, 10000))
# add_two_arrays(array_1, array_2)



import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Literal

from google.cloud import storage
from google.cloud.exceptions import NotFound
from rich.logging import RichHandler

from common_utils.cloud.base import GCPConnector

# Setup logging
logging.basicConfig(
    level="INFO",
    format="%(asctime)s [%(levelname)s]: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()],
)

logger = logging.getLogger("rich")


@dataclass
class GCS(GCPConnector):
    bucket_name: str
    storage_client: storage.Client = field(init=False, repr=False)
    bucket: storage.Bucket = field(init=False, repr=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.storage_client = storage.Client(
            credentials=self.credentials, project=self.project_id
        )
        self._init_bucket(self.bucket_name)

    def _init_bucket(self, bucket_name: str) -> None:
        """
        Initialize a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the GCS bucket.

        Returns
        -------
        None
        """
        self.bucket = self.storage_client.bucket(bucket_name)

    def check_if_bucket_exists(self) -> Literal[True, False]:
        """Check if a bucket exists."""
        try:
            self.storage_client.get_bucket(self.bucket_name)
            logger.info(f"Bucket {self.bucket_name} already exists")
            return True
        except NotFound:
            logger.info(
                f"""Bucket {self.bucket_name} does not exist.
                Please create it using `create_bucket`."""
            )
            return False

    def create_bucket(self) -> None:
        """Creates a new GCS bucket if it doesn't exist."""
        self.storage_client.create_bucket(self.bucket_name)
        logger.info(f"Bucket {self.bucket_name} created")

    def list_gcs_files(self, prefix: str = "", **kwargs: Dict[str, Any]) -> List[str]:
        """
        List the files in a GCS bucket with an optional prefix.

        Parameters
        ----------
        bucket_name : str
            The name of the GCS bucket.
        prefix : str, optional, default: ""
            The prefix to filter files in the bucket, by default "".
        **kwargs : Dict[str, Any]
            Additional arguments to pass to the list_blobs method.

        Returns
        -------
        gcs_files: List[str]
            The list of file names in the specified GCS bucket.
        """
        blobs = self.storage_client.list_blobs(
            self.bucket_name, prefix=prefix, **kwargs
        )
        gcs_files = [blob.name for blob in blobs]
        return gcs_files

    def create_blob(self, destination_blob_name: str) -> storage.Blob:
        """Creates a new blob in the bucket. See upload_blob for example."""
        blob = self.bucket.blob(destination_blob_name)
        return blob

    def upload_blob(
        self,
        source_file_name: str,
        destination_blob_name: str,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Uploads a file to a GCS bucket.
        https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-client-libraries

        # The ID of your GCS bucket
        # bucket_name = "your-bucket-name"
        # The path to your file to upload
        # source_file_name = "local/path/to/file"
        # The ID of your GCS object
        # destination_blob_name = "storage-object-name"

        Parameters
        ----------
        bucket_name : str
            The ID of your GCS bucket.
        source_file_name : str
            The path to your file to upload.
        destination_blob_name : str
            The ID of your GCS object.

        Returns
        -------
        None
        """
        # Optional: set a generation-match precondition to avoid potential race conditions
        # and data corruptions. The request to upload is aborted if the object's
        # generation number does not match your precondition. For a destination
        # object that does not yet exist, set the if_generation_match precondition to 0.
        # If the destination object already exists in your bucket, set instead a
        # generation-match precondition using its generation number.
        generation_match_precondition = 0

        blob = self.create_blob(destination_blob_name)

        blob.upload_from_filename(
            source_file_name,
            if_generation_match=generation_match_precondition,
            **kwargs,
        )

    def upload_directory(
        self, source_dir: str, destination_dir: str, **kwargs: Dict[str, Any]
    ) -> None:
        """
        Uploads a directory to a GCS bucket.
        https://cloud.google.com/storage/docs/uploading-objects
        """
        for file_path in Path(source_dir).glob("**/*"):
            if file_path.is_file():
                destination_blob_name = (
                    destination_dir + "/" + str(file_path.relative_to(source_dir))
                )
                self.upload_blob(str(file_path), destination_blob_name, **kwargs)

    def download_blob(
        self,
        source_blob_name: str,
        destination_file_name: str,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Downloads a blob from the bucket.
        https://cloud.google.com/storage/docs/downloading-objects

        # The ID of your GCS bucket
        # bucket_name = "your-bucket-name"

        # The ID of your GCS object
        # source_blob_name = "storage-object-name"

        # The path to which the file should be downloaded
        # destination_file_name = "local/path/to/file"
        """

        # Construct a client side representation of a blob.
        # Note `Bucket.blob` differs from `Bucket.get_blob` as it doesn't retrieve
        # any content from Google Cloud Storage. As we don't need additional data,
        # using `Bucket.blob` is preferred here.
        blob = self.create_blob(source_blob_name)

        blob.download_to_filename(destination_file_name, **kwargs)



import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import pandas as pd
from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from rich.logging import RichHandler

from common_utils.cloud.base import GCPConnector

# Setup logging
logging.basicConfig(
    level="INFO",
    format="%(asctime)s [%(levelname)s]: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()],
)

logger = logging.getLogger("rich")


@dataclass
class BigQuery(GCPConnector):
    # usually you call project_id.dataset_id.table_name
    dataset: str  # The ID of the dataset to use.
    table_name: str  # The name of the table to use.
    _dataset_id: str = field(init=False)  # The ID of the dataset to use.
    _table_id: str = field(init=False)  # The full ID of the table to use.
    bigquery_client: bigquery.Client = field(init=False)

    def __post_init__(self) -> None:
        """
        Initialize the BigQuery client for the GCP services.
        """
        super().__post_init__()
        self.bigquery_client = bigquery.Client(
            credentials=self.credentials, project=self.project_id
        )

    @property
    def table_id(self) -> str:
        """Return the full ID of the table to use."""
        return f"{self.project_id}.{self.dataset}.{self.table_name}"

    @property
    def dataset_id(self) -> str:
        """Return the ID of the dataset to use."""
        return f"{self.project_id}.{self.dataset}"

    def check_if_dataset_exists(self) -> Literal[True, False]:
        """Check if a dataset exists."""
        try:
            self.bigquery_client.get_dataset(self.dataset_id)
            logger.info(f"Dataset {self.dataset_id} already exists")
            return True
        except NotFound:
            logger.warning(
                f"""Dataset {self.dataset_id} does not exist.
                Please create it using `create_dataset`."""
            )
            return False

    def create_dataset(self) -> None:
        """Creates a new BigQuery dataset if it doesn't exist."""
        dataset = bigquery.Dataset(self.dataset_id)
        self.bigquery_client.create_dataset(dataset)
        logger.info(f"Created dataset {self.dataset_id}")

    def check_if_table_exists(self) -> Literal[True, False]:
        """Check if a table exists."""
        try:
            self.bigquery_client.get_table(self.table_id)
            logger.info(f"Table {self.table_id} already exists")
            return True
        except NotFound:
            logger.warning(
                f"""Table {self.table_id} does not exist.
                Please create it using `create_table`."""
            )
            return False

    def create_table(self, schema: List[bigquery.SchemaField]) -> None:
        """Creates a new BigQuery table if it doesn't exist."""
        table = bigquery.Table(self.table_id, schema=schema)
        self.bigquery_client.create_table(table)
        logger.info(f"Created table {self.table_id}")

    def query(
        self, query: str, as_dataframe: bool = True
    ) -> Union[List[Tuple[Any]], pd.DataFrame]:
        """
        Execute a query in BigQuery and return the result as a DataFrame.

        Parameters
        ----------
        query : str
            The SQL query to execute in BigQuery.

        Returns
        -------
        pd.DataFrame
            The result of the query as a DataFrame.
        """
        query_job = self.bigquery_client.query(query)
        results = query_job.result()
        return results.to_dataframe() if as_dataframe else results

    def load_job_config(
        self,
        schema: Optional[List[bigquery.SchemaField]] = None,
        write_disposition: str = "WRITE_APPEND",
        **kwargs: Dict[str, Any],
    ) -> bigquery.LoadJobConfig:
        """
        Creates a load job configuration for BigQuery.
        """
        return bigquery.LoadJobConfig(
            schema=schema, write_disposition=write_disposition, **kwargs
        )

    def load_table_from_dataframe(
        self,
        df: pd.DataFrame,
        job_config: bigquery.LoadJobConfig,
        **kwargs,
    ) -> None:
        """
        Loads data from a DataFrame to BigQuery.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to load.
        table_id : str
            The full ID of the table where the data will be loaded.
        schema: List[SchemaField]
            The schema fields to use for the table. None if the table already exists.
        """
        load_job = self.bigquery_client.load_table_from_dataframe(
            df, self.table_id, job_config=job_config, **kwargs
        )

        load_job.result()  # Waits for the job to complete.

        logger.info(f"Loaded {load_job.output_rows} rows into {self.table_id}.")



"""Logger class for logging to console and file."""
import logging
import os
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from rich.logging import RichHandler


class CustomFormatter(logging.Formatter):
    """This class overrides logging.Formatter's pathname to be relative path."""

    def format(self, record):
        record.pathname = os.path.relpath(record.pathname)
        return super().format(record)


@dataclass
class Logger:
    """
    Class for logger. Consider using singleton design to maintain only one
    instance of logger (i.e., shared logger).

    This logger implementation adheres to several principles of good
    software design:

    1. Single Responsibility Principle (SRP): The logger's responsibility
       is clear - managing and configuring the logging process. It creates
       logging handlers, formatters, and manages logging files and paths.

    2. Open-Closed Principle (OCP): The logger is open for extension (e.g.,
       one can easily extend it to add different types of handlers) but
       closed for modification (adding new functionality doesn't require
       modification of the existing code).

    3. Liskov Substitution Principle (LSP): This principle is not explicitly
       implemented in this class since it doesn't have any subclasses.

    4. Interface Segregation Principle (ISP): The Logger class does not
       depend on any interfaces it doesn't use. The class itself can be
       seen as a high-level interface for logging operations.

    5. Dependency Inversion Principle (DIP): The logger depends on
       abstractions (e.g., the logging module's Handler and Formatter
       classes), not concrete classes.

    1. Factory Pattern: In terms of design patterns, this Logger class follows the 'Factory'
       pattern by creating and returning logging handlers and formatters.
       For more justification, just realize that factory pattern is a creational pattern
       which provides an interface for creating objects in a superclass, but
       allows subclasses to alter the type of objects that will be created.

       For more info, see my design pattern notes.

    Areas for improvement:

    1. Consider implementing a Singleton pattern to ensure only one
       instance of Logger is used throughout the application.

    2. Consider adding thread safety to ensure that logs from different
       threads don't interfere with each other.

    3. The logger could be further extended to support other types of
       logging, such as sending logs to an HTTP endpoint.


    Example:
        Logger(
            log_file="pipeline_training.log",
            module_name=__name__,
            level=logging.INFO,
            propagate=False,
            log_root_dir="/home/user/logs",
        )

        --> produces the below tree structure, note the log_root_dir is the root
            while the session_log_dir is the directory for the current session.

        /home/
        │
        └───user/
            │
            └───logs/                        # This is the log_root_dir
                │
                └───2023-06-14T10:20:30/     # This is the session_log_dir
                    │
                    └───pipeline_training.log


    Parameters
    ----------
    log_file : Optional[str], default=None
        The name of the log file. Logs will be written to this file if specified.
        It must be specified if `log_root_dir` is specified.
    module_name : Optional[str], default=None
        The name of the module. This is useful for multi-module logging.
    level : int, default=logging.INFO
        The level of logging.
    propagate : bool, default=False
        Whether to propagate the log message to parent loggers.
    log_root_dir : Optional[str], default=None
        The root directory for all logs. If specified, a subdirectory will be
        created in this directory for each logging session, and the log file will
        be created in the subdirectory. Must be specified if `log_file` is specified.

    Attributes
    ----------
    session_log_dir : Optional[Union[str, Path]]
        The directory for the current logging session. This is a subdirectory
        within `log_root_dir` that is named with the timestamp of when the logger
        was created.
    logger : logging.Logger
        The logger instance.

    Raises
    ------
    AssertionError
        Both `log_file` and `log_root_dir` must be provided, or neither should be provided.
    """

    log_file: Optional[str] = None
    module_name: Optional[str] = None
    level: int = logging.INFO
    propagate: bool = False
    log_root_dir: Optional[str] = None

    session_log_dir: Optional[Union[str, Path]] = field(default=None, init=False)
    logger: logging.Logger = field(init=False)

    def __post_init__(self) -> None:
        assert (self.log_file is None and self.log_root_dir is None) or (
            self.log_file is not None and self.log_root_dir is not None
        ), "Both log_file and log_root_dir must be provided, or neither should be provided."

        self.logger = self._init_logger()

    def _create_log_output_dir(self) -> Path:
        current_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        session_log_dir = Path(self.log_root_dir) / current_time

        Path(session_log_dir).mkdir(parents=True, exist_ok=True)

        return session_log_dir

    def _get_log_file_path(self) -> Optional[Path]:
        if self.log_root_dir is not None:
            self.session_log_dir = self._create_log_output_dir()
            return self.session_log_dir / self.log_file
        return None

    def _create_stream_handler(self) -> RichHandler:
        stream_handler = RichHandler(rich_tracebacks=True, level=self.level)
        stream_handler.setFormatter(
            CustomFormatter(
                "%(asctime)s [%(levelname)s] %(pathname)s %(funcName)s L%(lineno)d: %(message)s",
                "%Y-%m-%d %H:%M:%S",
            )
        )
        return stream_handler

    def _create_file_handler(self, log_file_path: Path) -> logging.FileHandler:
        file_handler = logging.FileHandler(filename=str(log_file_path))
        file_handler.setFormatter(
            CustomFormatter(
                "%(asctime)s [%(levelname)s] %(pathname)s %(funcName)s L%(lineno)d: %(message)s",
                "%Y-%m-%d %H:%M:%S",
            )
        )
        return file_handler

    def _init_logger(self) -> logging.Logger:
        # get module name, useful for multi-module logging
        logger = logging.getLogger(self.module_name or __name__)
        logger.setLevel(self.level)

        logger.addHandler(self._create_stream_handler())

        log_file_path = self._get_log_file_path()
        if log_file_path:
            logger.addHandler(self._create_file_handler(log_file_path))
        logger.propagate = self.propagate
        return logger
